(*

  This fle implements n-gram grammars.

*)



(*
  SETTING UP PARAMETERS
*)

let usage () = print_endline (
    "\nBASIC USAGE: gram [subcommand] [infile]\n\n"^
      "Subcommands are:\n\n"^
      "find\t\t"^
      "accepts\t\t"^
      "2dot\n\n"^
      "See the documentation for more information.\n")


let cmd_list = List.tl (Array.to_list Sys.argv)
  
let error = ref false
let count = ref false
let wd = ref ""
let num = ref 2     (* default is bigram *)
let wb = ref "#"
let f1 = ref ""
let f2 = ref ""


let rec process_args list = 
  match list with
    | [] -> ()

    | "-n"::(n::t) -> num := (int_of_string n); process_args t

    | "-d"::(seg_delim::t) -> wd := seg_delim; process_args t

    | "-b"::(wd_bound::t) -> wb := wd_bound; process_args t

    | "-c"::t -> count := true; process_args t

    | "-d"::[]
    | "-b"::[]
    | "-n"::[]  -> error := true

    | arg::t ->
	match (!f1,!f2) with
	  | ("","") -> f1 := arg; process_args t
	  | (_,"") -> f2 := arg; process_args t
	  | _ -> error := true;;

(* note that f2 is a nonempty string iff f1 is a nonempty string *)

process_args (List.tl cmd_list);;

(* print_endline ("Err:"^(string_of_bool !error));; *)
(* print_endline ("Wd:"^(!wd));; *)
(* print_endline ("F1:"^(!f1));; *)
(* print_endline ("F2:"^(!f2));; *)
(* print_endline ("num:"^(string_of_int !num));; *)

module WD =
struct
  let delim = !wd 
  let lb = ""
  let rb = ""
end

(* Vocab is a module in FSA library.*)
module Gram = Xlist.Make(Delim.Comma_prs)(Vocab)
module VSet = Xset.Make(Delim.Comma_cbs)(Vocab)
module GramSet = Xset.Make(Delim.Semi_cbs)(Gram)
(*module Word = Xlist.Make(Delim.None)(Vocab) *)
module Word = Xlist.Make(WD)(Vocab)
module WordSet = Xset.Make(Delim.Newline_no_bs)(Word)


exception Ngram_exceeds_length

let get_ngram n word pos =  
  (* returns an n-gram from position pos in a word  *)
  let rec builder p m = 
    if m = 0 then Gram.empty 
    else
      try
	let g = Word.nth word p in
	Gram.build g (builder (p+1) (m-1))
      with 
	| Failure "nth" -> raise Ngram_exceeds_length 
  in
  builder pos n;;   

let add_ngrams_to_set n word set =
  (* adds all the n-grams in word to set *)
  let rec finder pos s =
    try
      let ng = get_ngram n word pos in
      if ng = Gram.empty
      then failwith 
	("Empty Ngram: "^(Gram.to_string word))
      else
      finder (pos+1) (GramSet.add ng s)
    with 
      |	Ngram_exceeds_length -> s
  in
  finder 0 set;;


let extract_ngrams n word =   
  (* returns a set of n-grams of size n from a word *)
  add_ngrams_to_set n word GramSet.empty


let accepts n g w = GramSet.subset (extract_ngrams n w) g


(* we use maps for counting ngrams *)

module GramMap = Map.Make(Gram)
  
(* type 'a map = NgramMap.t  *)

let increase_freq ngram map = 
    try 
      let f = GramMap.find ngram map in
      GramMap.add ngram (f +. 1.) map
    with
      | Not_found -> 
	  GramMap.add ngram 1. map


let add_ngrams_to_map n word map = 
  (* adds the ngrams in word to map *)
  let rec finder pos m =
    try
      let ng = get_ngram n word pos in
      finder (pos+1) (increase_freq ng m)
    with 
      |	Ngram_exceeds_length -> m
  in
  finder 0 map;;



let wboundary = Vocab.of_string !wb

let padwb n w = 
  let rec helper num wd= 
    match num with
      | 1 -> wd
      | _ -> helper (num-1) (Word.build wboundary wd)
  in
  helper n (Word.append w (Word.of_string !wb))

let count_ngrams n wordset = 
  (* returns a map of grams -> frequency *)
  WordSet.fold (fun w m -> 
    add_ngrams_to_map n (padwb n w) m)
    wordset GramMap.empty

let find_ngrams n wordset =
  (* returns a set of grams *)
  WordSet.fold (fun w m -> 
    add_ngrams_to_set n (padwb n w) m)
    wordset GramSet.empty

let count_ngrams_in_file n file = 
  let ic = open_in file in
  File.fold
    (fun s m -> add_ngrams_to_map n (padwb n (Word.of_string s)) m)
    ic GramMap.empty

let find_ngrams_in_file n file = 
  let ic = open_in file in
  File.fold
    (fun s m -> add_ngrams_to_set n (padwb n (Word.of_string s)) m)
    ic GramSet.empty


let print_map m = 
  Mmap.print m GramMap.iter "\t" "\n" 
    Gram.to_string string_of_float

let print_set gs = GramSet.iter 
  (fun g -> Gram.print g) gs

(* Making the FSA *)
module Symbol = Symbols.Make(Vocab)
module NodeSet = Xset.Make(Delim.Point_abs)(Node)
module NodeSetSet = Xset.Make(Delim.Newline_no_bs)(NodeSet)
module Edge = Edges.Make(Node)(Symbol)(Delim.Colon_no_bs)
module EdgeSet = Xset.Make(Delim.Newline_no_bs)(Edge)

(* Now we can make the FSA module. *)
module FSA = Acceptors.Make(Vocab)(Word)(WordSet)(Symbol)
  (Node)(NodeSet)(NodeSetSet)(Edge)(EdgeSet)(Delim.Excl_no_bs);;


exception Not_in_grammar

let find_sigma gs =
  VSet.remove wboundary
    (GramSet.fold
      (fun g1 vs1 -> 
	VSet.union
	  (Gram.fold_left
	    (fun vs2 v -> VSet.add v vs2)
	    vs1 g1
	  ) vs1
      ) gs VSet.empty)
    
let terminus_node origin_node a g =
  let gram = Gram.of_string (Node.to_string origin_node)
  in
  let new_gram = 
    Gram.rev (Gram.build a (Gram.rev gram))
  in
  if GramSet.mem new_gram g 
  then Node.of_string (Gram.to_string (Gram.tl new_gram))
  else raise Not_in_grammar

let rec is_final gram =
  if gram = Gram.empty then false
  else if Gram.hd gram = wboundary && Gram.tl gram = Gram.empty 
  then true
  else is_final (Gram.tl gram)

let to_fsa n g =
  let sigma = find_sigma g in
  let rec make_start = function
    | 0 -> Gram.empty
    | x -> Gram.build wboundary (make_start (x-1))
  in
  let start_gram = make_start (n-1) in
  let start = Node.of_string (Gram.to_string start_gram)
  in
  let rec make_delta nodes_to_do finished_nodes edgeset =
    if NodeSet.subset nodes_to_do finished_nodes then edgeset
    else
      let new_edges,new_nodes_to_do =
	NodeSet.fold
	  (fun node (es1,ns1) ->
	    let new_edges2,new_nodes2 = 
	      VSet.fold
		(fun v (es2,ns2) ->
		  try
		    let new_node = terminus_node node v g
		    in
		    let new_edge = 
		      Edge.make node new_node (Symbol.of_x v)
		    in
		    (EdgeSet.add new_edge es2, NodeSet.add new_node ns2)
		  with Not_in_grammar -> (es2,ns2)
		) 
		sigma (es1,ns1)
	    in
	    (EdgeSet.union es1 new_edges2, NodeSet.union ns1 new_nodes2)
	  ) 
	  nodes_to_do (EdgeSet.empty,NodeSet.empty)
      in
      make_delta 
	new_nodes_to_do 
	(NodeSet.union finished_nodes nodes_to_do)
	(EdgeSet.union new_edges edgeset)
  in
  let edges = make_delta (NodeSet.singleton start) 
    NodeSet.empty EdgeSet.empty
  in
  let finals = 
    GramSet.fold
      (fun gram ns -> NodeSet.add 
	(Node.of_string 
	  (Gram.to_string (Gram.rev (Gram.tl (Gram.rev gram))))) 
	ns)
      (GramSet.filter 
	(fun gram -> is_final gram)
	g
      )
      NodeSet.empty
  in
  FSA.make (NodeSet.singleton start) finals edges



(* MAIN PROGRAM BEGINS HERE *)

let raise_error () = 
  let () = List.iter (fun x -> print_string (x^":")) cmd_list in
  let () = print_endline "" in
  let () = print_endline "Unrecognized arguments" in 
  usage ();;

let main () =
  let subcommand = 
    try List.hd cmd_list 
    with _ -> print_endline "e1"; raise_error (); failwith ""
  in
  if !error then begin print_endline "e2"; raise_error () end
  else
    match subcommand with
      | "accepts" ->
	  begin
	    let g,w = 
	      match (!f1,!f2) with
(*		| ("","")   (PRSet.of_string (File.read ()), input_line stdin) *)
		| (_,"") -> raise_error (); failwith "" (*(PRSet.of_file a, input_line stdin)*)
		| (a,b) -> (GramSet.of_file a, b)
	    in
	    print_endline (string_of_bool (accepts !num g (Word.of_string w)))
	  end

    | "2fsa" ->
	begin
	  let g =
	    match (!f1,!f2) with
	      | ("","") -> File.of_string "temp_gram_grmr" (File.read ());
		  GramSet.of_file "temp_gram_grmr"
	      | (a,"") -> GramSet.of_file a
	      | _ -> raise_error (); failwith ""
	  in
	  FSA.print (to_fsa !num g)
	end

    | "find" ->
	begin
	  if !count then
	    match (!f1,!f2) with
	      | ("","") -> print_map (count_ngrams !num (WordSet.of_string (File.read ())))
	      | (a,"") -> print_map (count_ngrams_in_file !num a)
	      | _ -> print_endline "e3"; raise_error ()
	  else
	    match (!f1,!f2) with
	      | ("","") -> print_set (find_ngrams !num (WordSet.of_string (File.read ())))
	      | (a,"") -> print_set (find_ngrams_in_file !num a)
	      | _ -> print_endline "e3"; raise_error ()
	end

    | _ -> print_endline "e4"; raise_error ();;

main ();;
